
//*****************************************************************************************//
//                                      Includes                                           //
//*****************************************************************************************//

#include "includes.all"

//*****************************************************************************************//
//                                      Texture                                            //
//*****************************************************************************************//

//*****************************************************************************************//
//                          Private Packing/Unpacking Facilities                           //
//*****************************************************************************************//

#define redMask 0X000000FF
#define greenMask 0X0000FF00
#define blueMask 0X00FF0000
#define alphaMask 0XFF000000

#define redShift 0
#define greenShift 8
#define blueShift 16
#define alphaShift 24

inline long pack (const long red, const long green, const long blue, const long alpha) {
	return 
		((red << redShift) & redMask) | 
		((green << greenShift) & greenMask) | 
		((blue << blueShift) & blueMask)| 
		((alpha << alphaShift) & alphaMask);
}

inline long unpackRed (const long pixel) {
	return (pixel >> redShift) & 0X000000FF;
}

inline long unpackGreen (const long pixel) {
	return (pixel >> greenShift) & 0X000000FF;
}

inline long unpackBlue (const long pixel) {
	return (pixel >> blueShift) & 0X000000FF;
}

inline long unpackAlpha (const long pixel) {
	return (pixel >> alphaShift) & 0X000000FF;
}

inline bool isPowerOf2 (long value) {
	return (value & ~(value - 1)) == value;
}	

//*****************************************************************************************//
//                                Texture Implementation                                   //
//*****************************************************************************************//

Texture::Texture (long width, long height, TextureType type) {
	//Fill in all attributes.
	this->type = type;
	this->width = width;
	this->height = height;
	this->bytes = new long [width * height]; 
	this->textureHandle = -1; //If we don't get a handle, it will still be -1.
	this->textureName = NULL; //So far.
	glGenTextures (1, &this->textureHandle);
}

Texture::~Texture () {
	delete [] bytes; delete [] textureName;
}

Texture *Texture::readTexture (char *fileName) {
	//Read a texture and fill in all attributes; returns NULL if unsuccessful.

	//Get Microsoft to read it (they must know how).
    HBITMAP bitmapHandle = (HBITMAP) LoadImage (NULL, fileName, 
    	IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);	
	if (bitmapHandle == NULL) {log ("\nFile \"%s\" not found...", fileName); return NULL;}

    //Find out how big it is.
	BITMAP bitmap; GetObject (bitmapHandle, sizeof (bitmap), &bitmap);
	long width = bitmap.bmWidth;	
	long byteWidth = bitmap.bmWidthBytes; 
	long height = bitmap.bmHeight;
	long pixelSize = bitmap.bmBitsPixel;
	long planes = bitmap.bmPlanes;
	char *bits = (char *) bitmap.bmBits;

	if (!isPowerOf2 (width) || !isPowerOf2 (height)) {
		//OpenGL needs a power of 2 (it is possible to force OpenGL to resize it but that's slow).
		prompt ("\nBitmap \"%s\" is not a power of 2; width %d, height %d.", fileName, width, height);
		log ("\nBitmap \"%s\" is not a power of 2; width %d, height %d.", fileName, width, height);
	}

	//Allocate space for RGBA format.
	Texture *texture = new Texture (width, height, RGBAType);
	if (texture == NULL) {DeleteObject (bitmapHandle); return NULL;}

	//Prepare to transfer the bits.
	long *destination = texture->bytes; //REVERSES: destination += height * width; //past the end
    BYTE *source = (BYTE *) bits; //at the beginning

	//Move the bits from the bitmap handle to the texture.	 
	long red, green, blue; long alpha = 255; //0=>transparent, 255=>opaque
	if (pixelSize == 24) {
		//Successively read 3 RGB bytes at a time, write RGBA bytes. 
    	for	(long j = height; j > 0; j--) {
			//REVERSES: destination -= width; 
			long *to = destination; BYTE *from = source;
    		for (int i = 0; i < width; i++) { 
				blue = *from++;	green = *from++; red = *from++;
				*to++ = pack (red, green, blue, alpha);
    		}   
			destination += width; 											   
			source += byteWidth;
		} 
	} else if (pixelSize == 8) { 
		//Obtains the palette information from the handle.
		typedef PALETTEENTRY PaletteEntries [256];
		HDC deviceContextHandle = CreateCompatibleDC (NULL);
		SelectObject (deviceContextHandle, bitmapHandle);
		PaletteEntries paletteEntries; 
		ZeroMemory (&paletteEntries, sizeof (PaletteEntries));
		GetDIBColorTable (deviceContextHandle, 0, 256, (RGBQUAD *) &paletteEntries);
		DeleteDC (deviceContextHandle);	
		long packedPaletteEntries [256];

		//Pack it into a nice RGBA array.
		long *palette = (long *) &paletteEntries; 
		for (int i = 0; i < 256; i++) {
			//DIB color tables have their colors stored BGR not RGB (so flip).
			BYTE red = paletteEntries [i].peBlue;
			BYTE green = paletteEntries [i].peGreen;
			BYTE blue = paletteEntries [i].peRed;
			packedPaletteEntries [i] = pack (red, green, blue, alpha);
		} 

		//Move the bits.
    	for	(long j = height; j > 0; j--) {
			//REVERSES: destination -= width; 
			long *to = destination; BYTE *from = source; 
			for (int i = width; i > 0; i--) { 
				*to++ = packedPaletteEntries [*from++];
			}
			destination += width; 
      		source += byteWidth; 
    	} 
	} else {
		//We only handle 8 and 24 bits so far...
		log ("\nRead texture can only handle 8 and 24 bit textures, not %d.", pixelSize); 
		delete texture; DeleteObject (bitmapHandle); return NULL;
	}

	texture->textureName = new char [strlen (fileName) + 1];
	strcpy (texture->textureName, fileName);
    DeleteObject (bitmapHandle); return texture;
}

bool Texture::readTextureExtent (char *fileName, long &width, long &height) {
	//Read just enough of a texture to determine it's extent; i.e., width and height.
	//Returns true if successful; false otherwise.

	//Get Microsoft to read it (they must know how).
    HBITMAP bitmapHandle = (HBITMAP) LoadImage (NULL, fileName, 
    	IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);	
	if (bitmapHandle == NULL) {log ("\nFile \"%s\" not found...", fileName); return false;}

    //Find out how big it is.
	BITMAP bitmap; GetObject (bitmapHandle, sizeof (bitmap), &bitmap);
	width = bitmap.bmWidth;	
	height = bitmap.bmHeight;
	DeleteObject (bitmapHandle); 
	return true;
}

void Texture::activate () {
	if (textureHandle == -1) {glDisable (GL_TEXTURE_2D); return;}
	glEnable (GL_TEXTURE_2D);  //Turn on texturing.
	glBindTexture (GL_TEXTURE_2D, textureHandle); //Bind the current texture.
}

void Texture::load (bool mipmapping, bool forceClamp) {
	//Give the texture to the game card.
	if (textureHandle == -1 || bytes == NULL) {
		log ("\nTexture handle is %d (-1 means not set), bytes %x (null means not read)...", textureHandle, bytes);
		return;
	}
	static long alignment [2] = {4, 1};
	static GLint wrap [2] = {GL_REPEAT, GL_REPEAT}; //GL_CLAMP is currently not used
	static GLint components [2] = {GL_RGBA8, GL_RGB8};
	static GLenum format [2] = {GL_RGBA, GL_RGB};

	activate ();
	glPixelStorei (GL_PACK_ALIGNMENT, alignment [type]);
	GLint wrapping = forceClamp ? GL_CLAMP : wrap [type];
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrapping);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrapping);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	GLint minFilter = mipmapping ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR;
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
	if (mipmapping)
		gluBuild2DMipmaps (GL_TEXTURE_2D, components [type], width, height,
			format [type], GL_UNSIGNED_BYTE, bytes);
	else
		glTexImage2D (GL_TEXTURE_2D, 0, components [type], width, height, 0,
			format [type], GL_UNSIGNED_BYTE, bytes);
	log ("\nLoad \"%s\", handle %d.", textureName, textureHandle);
}

void Texture::unload () {
	log ("\nUnload \"%s\", handle %d.", textureName, textureHandle);
	if (textureHandle != -1) glDeleteTextures (1, &textureHandle);
	textureHandle = -1;
}
