
//*****************************************************************************************//
//                                      Includes                                           //
//*****************************************************************************************//

#include "includes.all"

//*****************************************************************************************//
//                            Worldcraft Reader Discussion                                 //
//*****************************************************************************************//

//Worldcraft provides data that consists essentially of planes which define a convex region.
//By constructing these planes and using then to cut each other, we end up computing the values of
//coordinates that can be used to construct the faces. It does not give us the face coordinates.
//
//Worldcraft also provides data that allows use to compute texture coordinates given that
//we have face coordinates. It does not give us texture coordinates.

//To construct, a suitable transformation, we need to understand what Worldcraft had in mind.
//Essentially, they wanted to take their normal coordinates and transform them to the appropriate
//texture coordinate by writing 
//
//					normalCoordinate * M = textureCoordinate.
//
//But Worldcraft texture coordinates are bitmap size DEPENDENT (i.e., coordinates can run from
//0 to the width-1 in the x-direction and 0 to height-1 in the y-direction). By contrast, OpenGL
//is bitmap size INDEPENDENT (i.e., coordinates ranges from 0.0 to 1.0 no matter what the size
//of the bitmap). 
//
//Let N (for NORMALIZE) be the transformation that scales the extent (width and height) of a bitmap 
//down to range 0.0 to 1.0. Then the transformation we need looks like
//
//					normalCoordinate * M * N = textureCoordinate.
//
//Now, in general a transformation consists of three parts: a rotation R, a scale S, and a
//translation T. So the transformation can actually be written as
//
//					normalCoordinate * R * S * T * N = textureCoordinate.
//
//In general, it's important to scale before you translate because scaling moves points. So 
//if you scale first, then translation can move the point from where it happens to be to where
//you want it. If you translate first, you will have to translate it to a spot which gets
//it it to where you really want it to be after you scale it. It can be done this way but it's
//harder to control.
//
//Now for some esoteric knowledge. Whatever you do to normal coordinates, you must do the
//INVERSE to texture coordinates. This could be stated as a theorem but that wouldn't make
//it any clearer. Consider an example instead, a rectangle displaying the left half of a texture.
//
//		Rectangle:										Texture
//
//		   tx@ty=0.0@1.0			tx@ty=0.5@1.0		
//			  .____________________________.        .____________________________.
//			  .                            .        .                            .
//			  .     a       a      a       .        .  a  a  a         b  b  b   .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .     a       a      a       .        .  a  a  a         b  b  b   .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .     a       a      a       .        .  a  a  a         b  b  b  .
//			  .____________________________.        .____________________________.
//
//		   tx@ty=0.0@0.0			tx@ty=0.5@0.0		
//
//It doesn't matter what the normal coordinates of the rectangle are. But since the texture
//coordinates have a x part (tx) that goes from 0.0 to 0.5, what you will see drawn in the
//rectangle are the left half of the texture; i.e., JUST THE A's (actually twice as big as 
//the original A's but I couldn't draw bigger A's with the keyboard). 	
//
//Now, let's translate the texture coordinates by 0.25 to see what happens. We should get
//the following:
//
//		   tx@ty=0.25@1.0			tx@ty=0.75@1.0		
//			  .____________________________.        ._______|______|_____|_______.
//			  .                            .        .                            .
//			  .     a              b       .        .  a  a  a         b  b  b   .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .     a              b       .        .  a  a  a         b  b  b   .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .     a              b       .        .  a  a  a         b  b  b  .
//			  .____________________________.        .____________________________.
//
//		   tx@ty=0.25@0.0			tx@ty=0.75@0.0		
//
//Since the x part (tx) of range from 0.25 to 0.75, we see the rightmost column of A's
//and the leftmost column of B's. Now, let just translate the texture coordinates by 0.25
//one more time. After the translation, the texture coordinates are
//
//		   tx@ty=0.5@1.0			tx@ty=1.0@1.0		
//			  .____________________________.        .____________________________.
//			  .                            .        .                            .
//			  .     b       b      b       .        .  a  a  a         b  b  b   .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .     b       b      b       .        .  a  a  a         b  b  b   .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .     b       b      b       .        .  a  a  a         b  b  b  .
//			  .____________________________.        .____________________________.
//
//		   tx@ty=0.5@0.0			tx@ty=1.0@0.0	
//
//What you see first are the A's, then half A's and half B's, then just B's. You see
//the texture MOVING LEFT. THEOREM: TRANSLATING TEXTURE COORDINATE RIGHT CAUSES THE
//TEXTURE TO MOVE LEFT.
//
//I'll show one more example, this time in the context of scaling. Consider
//
//
//		   tx@ty=0.0@0.33			tx@ty=0.33@0.33		      
//			  .____________________________.        .____________________________.
//			  .                            .        .                            .
//			  .                            .        .  C   C   C    C            .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .      A           A         .        .  B   B   B    C            .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .                            .        .  A   A   B    C            .
//			  .____________________________.        .____________________________.
//
//		   tx@ty=0.0@0.0			tx@ty=0.33@0.0		
//
//With those texture coordinate, all you see are large A's. Let's scale the texture by 2.0. 
//Now you should see
//
//		   tx@ty=0.0@0.66			tx@ty=0.66@0.66		      
//			  .____________________________.        .____________________________.
//			  .                            .        .                            .
//			  .                            .        .  C   C   C    C            .
//			  .      B       B      B      .        .                            .
//			  .                            .        .                            .
//			  .                            .        .  B   B   B    C            .
//			  .                            .        .                            .
//			  .      A       A      B      .        .                            .
//			  .                            .        .  A   A   B    C            .
//			  .____________________________.        .____________________________.
//
//		   tx@ty=0.0@0.0			tx@ty=0.66@0.0		
//
//Scaling by 1.5 again gives us
//
//		   tx@ty=0.0@0.99			tx@ty=0.99@0.99		      
//			  .____________________________.        .____________________________.
//			  .                            .        .                            .
//			  .  C   C   C    C            .        .  C   C   C    C            .
//			  .                          .        .                            .
//			  .                            .        .                            .
//			  .  B   B   B    C            .        .  B   B   B    C            .
//			  .                            .        .                            .
//			  .                            .        .                            .
//			  .  A   A   B    C            .        .  A   A   B    C            .
//			  .____________________________.        .____________________________.
//
//		   tx@ty=0.0@0.0			tx@ty=0.99@0.0		
//
//What do you see visually? You see big A's, which shrink down until you see A' with B's, which
//shrink down further until you see A's with B's with C'. THEOREM: SCALING UP CAUSES THE
//TEXTURE TO SCALE DOWN.
//
//The moral of this story should be clear. Since the Worldcraft transformation R * S * T * N
//is actually being applied to textures, the inverse is needed. More specifically,
//
//					(R * S * T * N)-1				Notation: let M-1 means M inverse
//
//There is a well known theorem that I could prove that states "THE INVERSE OF A PRODUCT
//IS THE PRODUCT OF THE INVERSES WITH THE ELEMENT ORDER SWITCHED); i.e., the inverse of the
//above is 
//
//					N-1 * T-1 * S -1 * R-1
//
//I tried this transformation and it DID NOT WORK; i.e., what you saw when I rendered
//a polygon what NOT what you say in Worldcraft. So what I had to do is guess what 
//Worldcraft might be doing and do the same thing. Here's my guess. 
//
//I believe the Worldcraft implementers were not aware how inverses worked. They tried
//to use the original transformation; i.e.,
//	
//					R * S * T * N
//
//But this didn't work. When they rotated right in their editor, the texture rotated left
//(so they negated the sign of the rotation), when they translated right, the texture went
//left (so they negated the sign of the translation), when they scaled up, the texture
//shrunk (so they replaced the scale factor by 1/scale). Each of these fudges is replacing
//the original desire by the inverse. So they ended up with the following transformation:
//	
//					R-1 * S-1 * T-1 * N-1
//
//This means that what you see in Worldcraft is the result of applying this transformation
//to the normal textures. If we want to see the same thing, we need to do it this way too.

//Now Worldcraft doesn't provide one matrix or even the individual transformation matrices 
//directly (that would be too easy). Instead, they provide a number of independent pieces 
//that can be used to construct this. In particular, they provide 2 of the axes for the 
//rotation matrix (the new xAxis and new yAxis; we can compute the missing third zAxis). 
//We also have to do some work to construct a rotation matrix out of axis information.
//They also provide the scale and translation information as corresponding x- and y-components.
//To compute N, we also need to know the extent of the texture even though we don't
//need to render it in the reader. That means we have to read the file just to determine
//the extent.

//One final thing. Worldcraft uses it's own units; there are 8 worldcraft units to a foot.
//We would like our engine to use 1 unit to means something real; how about 1 meter?
//So we apply this final conversion worldcraftToMeters) to get that.

//*****************************************************************************************//
//                              Worldcraft Reader Terminology                              //
//*****************************************************************************************//

//The reader reads in raw worldcraft data that describes objects (brushes they call them) in 
//terms of planes (that we'll call a worldcraft side plane since it is not at all what a 
//standard plane looks like). The reader reads all this information in, converts the side 
//plane data into coordinates for faces and returns a vector of worldcraft objects (a map 
//that contains  everything worldcraft provided as key/value pairs (both keys and values are
//strings) except for one additional key:

//*****************************************************************************************//
//                      Utilities That Could Have Been Placed in Plane                     //
//*****************************************************************************************//

//But we didn't because there are so specialized that no one else would use them.

//Forward references...
bool isCounterClockwise (Plane &plane, Point &point1, Point &point2, Point &point3);
PointCollection *anyFarCornerPoints (Plane &plane);
PointCollection *clip (Plane &plane, PointCollection *candidatePoints);
Point whereIsIntersectionOfPlaneCrossingLine (Plane &plane, Point &fromPoint, Point &toPoint);
Point binarySearchForCrossingPoint (Plane &plane, Point &positivePoint, Point &negativePoint);

bool isCounterClockwise (Plane &plane, Point &point1, Point &point2, Point &point3) {
	//Are these three points on the plane in counter-clockwise order WITH RESPECT TO THE POSITIVE SIDE OF THIS PLANE?

	//Construct a temporary plane and see if the normal is pointing in the same direction.
	//If no such normal can be constructed, return false; (true would also be just as valid).
	Vector unnormalizedNormal; double squaredLength;
	if (!plane.normalIsValid (point1, point2, point3, unnormalizedNormal, squaredLength)) return false;
	Vector normalizedNormal = unnormalizedNormal * (1.0 / sqrt (squaredLength));
	return (normalizedNormal - plane.normal).squaredLength () < epsilon (); //i.e., new normal == plane.normal
}

PointCollection *anyFarCornerPoints (Plane &plane) {
    //Compute 4 points that form a rectangle in the plane ensuring that they are provided in 
	//counter-clockwise order.
    //	Approach: If we find which coordinate of the normal is maximal ignoring sign (call it r which is 
	//	either x, y, or z), then the other two can be set to maximal values and then we can
	//	solve for r to determine where on the plane it lies.

	//The equation of the plane is aX+bY+cZ+d = 0 where a=normal.x = x (for short), 
	//b=normal.y = y (for short), c=normal.z = z (for short), d = minusP0Dotnormal.
	//					i.e., the equation is xX+yY+zZ+d = 0.
	//When you solve for X, for example, you get X = (-d-yY-zZ)/x
	//Since every solution has -d in it, it will be convenient to negate d once and for all
	//so that the solution X is written (d-yY-zZ)/x = (d-Yy-Zz)/x.
	//Plugging in values like Y=p, z=m gives (d-py-mz)/x
    double x = plane.normal.x; double y = plane.normal.y; double z = plane.normal.z;
    double d = -plane.minusP0DotNormal; //To avoid using -d everywhere below.
    double p = 1.0e10; double m = -p; //p=plus and m=minus

	//Use short forms p1,p2,p3,p4 to shorten the code below.
	Point *p1 = new Point; Point *p2 = new Point; Point *p3 = new Point; Point *p4 = new Point; 

    if (absolute (x) > absolute (y)) {
		if (absolute (x) > absolute (z))  {
			//Solve for x given y,z components which are {p,p}, {p,m}, {m,m}, {m,p}.
			Point point1 = {(d - (p * y) - (p * z)) / x, p, p}; //e.g., (d-py-mz)/x from above.
			Point point2 = {(d - (p * y) - (m * z)) / x, p, m};
			Point point3 = {(d - (m * y) - (m * z)) / x, m, m};
			Point point4 = {(d - (m * y) - (p * z)) / x, m, p};
			*p1 = point1; *p2 = point2; *p3 = point3; *p4 = point4;
		} else {
			//Solve for z given x,y components which are {p,p}, {p,m}, {m,m}, {m,p}.
			Point point1 = {p, p, (d - (p * x) - (p * y)) / z};
			Point point2 = {p, m, (d - (p * x) - (m * y)) / z};
			Point point3 = {m, m, (d - (m * x) - (m * y)) / z};
			Point point4 = {m, p, (d - (m * x) - (p * y)) / z};
			*p1 = point1; *p2 = point2; *p3 = point3; *p4 = point4;
		}
	} else {
		if (absolute (y) > absolute (z)) {
			//Solve for y given x,z components which are {p,p}, {p,m}, {m,m}, {m,p}.
			Point point1 = {p, (d - (p * x) - (p * z)) / y, p};
			Point point2 = {p, (d - (p * x) - (m * z)) / y, m};
			Point point3 = {m, (d - (m * x) - (m * z)) / y, m};
			Point point4 = {m, (d - (m * x) - (p * z)) / y, p};
			*p1 = point1; *p2 = point2; *p3 = point3; *p4 = point4;
		} else {
			//Solve for z given x,y components which are {p,p}, {p,m}, {m,m}, {m,p}.
			Point point1 = {p, p, (d - (p * x) - (p * y)) / z};
			Point point2 = {p, m, (d - (p * x) - (m * y)) / z};
			Point point3 = {m, m, (d - (m * x) - (m * y)) / z};
			Point point4 = {m, p, (d - (m * x) - (p * y)) / z};
			*p1 = point1; *p2 = point2; *p3 = point3; *p4 = point4;
		}
	}
	
	PointCollection *points = new PointCollection;
	if (isCounterClockwise (plane, *p1, *p2, *p3)) {
		//Already counter-clockwise.
		points->push_back (p1); points->push_back (p2);
		points->push_back (p3); points->push_back (p4);
	} else {
		//Must reverse the order to get counter-clockwise.
		points->push_back (p4); points->push_back (p3);
		points->push_back (p2); points->push_back (p1);
	}
	return points;
}

PointCollection *clip (Plane &plane, PointCollection *candidatePoints) {
    //Returns clipped points; i.e., those on the positive or zero side (also eliminates zero length edges).
    if (candidatePoints->empty ()) return new PointCollection; //Nothing to clip.

    //Keep what remains after clipping.
    PointCollection *points = new PointCollection;
    for (long index = 0; index < candidatePoints->size (); index++) {
        Point *point1 = (*candidatePoints) [index];
		//Let point2 be the next point and wraparound to the beginning if at the end.
        Point *point2 = (*candidatePoints) [index == candidatePoints->size () - 1 ? 0 : index + 1]; 
        LineSign lineSign = plane.whereIsLine (*point1, *point2);
//log ("\n   Consider point [%3.2f,%3.2f,%3.2f] -> [%3.2f,%3.2f,%3.2f].", point1->x, point1->y, point1->z,
//point2->x, point2->y, point2->z);

		switch (plane.whereIsLine (*point1, *point2)) {
			case frontSign: 
				points->push_back (point1->newCopy ()); //log (" ADD FIRST (front)");
				break;
			case backSign: //log (" DISCARD (back)");
				//Discard. 
				break;
			case onSign: 
				points->push_back (point1->newCopy ()); //log (" ADD FIRST (on)");
				break;
			case straddlesSign:
				PointSign point1Sign = plane.whereIsPoint (*point1);
				PointSign point2Sign = plane.whereIsPoint (*point2);

				if (point1Sign == positiveSign || point1Sign == zeroSign) {
					points->push_back (point1->newCopy ()); //log (" ADD FIRST (straddles, first point front or on)");
				}

				if (point1Sign == positiveSign && point2Sign == negativeSign) {
					Point intersection = whereIsIntersectionOfPlaneCrossingLine (plane, *point1, *point2);
					points->push_back (intersection.newCopy ()); //log (" ADD INT (straddle, intersection)");
				} else if (point1Sign == negativeSign && point2Sign == positiveSign) {
					Point intersection = whereIsIntersectionOfPlaneCrossingLine (plane, *point2, *point1);
					points->push_back (intersection.newCopy ()); //log (" ADD INT (straddle, intersection)");
				}
		}
	}
//log ("\nAfter first clip, have %d points.", points->size ());

    //Remove duplicates and useless points (less than 3 points means it can't be used to represent a face).
    PointCollection *trimmedPoints = new PointCollection;
    for (index = 0; index < points->size (); index++) {
        Point *previousPoint = (*points) [index == 0 ? points->size () - 1 : index - 1]; //wrap to right end
        Point *point = (*points) [index];
		if (point->squaredDistanceTo (*previousPoint) < squaredEpsilon ())
			; //Discard 
		else trimmedPoints->push_back (point->newCopy ());
	}
//log ("\nAfter trimming, have %d points.", trimmedPoints->size ());

	deletePointCollection (points);
	if (trimmedPoints->size () >= 3) return trimmedPoints;
	deletePointCollection (trimmedPoints); //deletes points too...
	return new PointCollection;
}

Point whereIsIntersectionOfPlaneCrossingLine (Plane &plane, Point &fromPoint, Point &toPoint) {
	//Note: this is a special intersection method that REQUIRES the points to cross the plane;
	//i.e., we insist that one point be on the positive side and one be on the negative side.
//log ("\nIntersect line through plane [%3.2f,%3.2f,%3.2f,%3.2f].", plane.normal.x, plane.normal.y, plane.normal.z, plane.minusP0DotNormal);
//log ("\nLine [%3.2f,%3.2f,%3.2f]->[%3.2f,%3.2f,%3.2f].", fromPoint.x, fromPoint.y, fromPoint.z, toPoint.x, toPoint.y, toPoint.z);
	Vector direction = toPoint - fromPoint; direction.normalize ();
	if (plane.normalNearlyPerpendicular (direction)) {
//log ("\nNearly perpendicular...");
		//Raw math doesn't work in such a situation.
		if (plane.whereIsPoint (fromPoint) == positiveSign) {
			if (plane.whereIsPoint (toPoint) != negativeSign) {halt ("\nPoints should cross the plane");}
			return binarySearchForCrossingPoint (plane, fromPoint, toPoint);
		} else {
			if (plane.whereIsPoint (toPoint) != positiveSign) {halt ("\nPoints should cross the plane");}
			return binarySearchForCrossingPoint (plane, toPoint, fromPoint);
		} 
	} else {
		//Use conventional math. More specifically, if N.p + d = 0 is the equation of the plane and
		//p = p0 + D*t for any t, then N.p0 + N.Dt + d = 0 and t = (-d - N.p0)/N.D. Solve for p.
		//Notice that 'aPlane distanceToPoint: p0 = N.p0 + d. So 
		//		t = -(aPlane distanceToPoint: p0) / N.D."
		double t = -plane.distanceToPoint (fromPoint) / (plane.normal.dot (direction));
//log ("\nNice solution at t = %3.2f/%3.2f = %3.2f to give [%3.2f,%3.2f,%3.2f].", -plane.distanceToPoint (fromPoint), plane.normal.dot (direction), t, fromPoint + direction * t);
		return fromPoint + direction * t;
	}
}

Point binarySearchForCrossingPoint (Plane &plane, Point &positivePoint, Point &negativePoint) {
    Point average = (positivePoint + negativePoint) * 0.5;
    PointSign averageSign = plane.whereIsPoint (average);
    if (averageSign == zeroSign) return average;
    if (averageSign == positiveSign)
        return binarySearchForCrossingPoint (plane, average, negativePoint);
    else return binarySearchForCrossingPoint (plane, positivePoint, average);
}


//*****************************************************************************************//
//                                   WorldcraftSidePlane                                   //
//*****************************************************************************************//

//The worldcraft side plane contains worldcraft data both as it came in and as computed.

struct WorldcraftSidePlane {
	Point point1; Point point2; Point point3;
	char *textureName;
	Point xAxis; double xAxisOffset; double xAxisScale;
	Point yAxis; double yAxisOffset; double yAxisScale;
	double rotation;
	//Computed from above.
	Plane plane;
	Transformation transformation;
};

declareCollection (WorldcraftSidePlane);

//*****************************************************************************************//
//                                    Forward References                                   //
//*****************************************************************************************//

WorldcraftGroupCollection *worldcraftReader (char *filename);
WorldcraftGroup *readRestOfWorldcraftGroup ();
char *readRestOfDoubleQuotedString ();
WorldcraftSidePlaneCollection *readRestOfSidePlanes ();
WorldcraftSidePlane *readRestOfSidePlane ();
char *readTextureName ();
FaceCollection *convertToFaces (WorldcraftSidePlaneCollection *sidePlanes);

//*****************************************************************************************//
//                            Worldcraft Reader Implementation                             //
//*****************************************************************************************//

#define add(key,value) insert (WorldcraftGroup::value_type (key, value)) //easier to read

static ::ifstream file; 
static char directoryName [500];
static long mapCounter = 0; //For debugging only.

WorldcraftGroupCollection *worldcraftReader (char *filename) {
log ("\nRead worldcraft file \"%s\".", filename);
	file.open (filename, ios::in); 
	if (file.fail ()) {halt ("\nFile \"%s\" could not be opened.", filename);}
log ("\nFile \"%s\" successfully opened.", filename);
	//Get the directory name.
	char drive [20], directory [500], name [100], extension [20];
	_splitpath (filename, drive, directory, name, extension);
	strcpy ((char *) &directoryName [0], (char *) &drive [0]);
	strcat ((char *) &directoryName [0], (char *) &directory [0]);
log ("\nThe directory is \"%s\".", directoryName);
	WorldcraftGroupCollection *objects = new WorldcraftGroupCollection;
	mapCounter = 0; //For debugging only.
	char next;
	while (file >> next) {//>> skips whitespace
//log ("\nRead character \"%c\".", next);
		if (next == '{') 
			objects->push_back (readRestOfWorldcraftGroup ());
		else quit ("Unexpected input \"%c\" in worldcraft file.", next);
	}
	file.close ();
	return objects;
}

WorldcraftGroup *readRestOfWorldcraftGroup () {
    //Reads and creates a worldcraft object; assumes leading '{' already discarded.

//log ("\n    Read rest of worldcraft object...");
	//Create an object and start to fill it in.
    WorldcraftGroup *group = new WorldcraftGroup; mapCounter++;
	ObjectCollection *objects = new ObjectCollection;
	group->add ("objects", objects);

    //Read what is to become contents for the group...
	char next;
	while (file >> next) {//>> skips whitespace
//log ("\n    Read character \"%c\".", next);
		switch (next) {
			case '"': {
				char *name = readRestOfDoubleQuotedString ();
				if (!(file >> next)) quit ("Unexpected error after reading a double quote.");
				if (next != '"') quit ("Expected another string after \"%s\".", name);
				void *value = readRestOfDoubleQuotedString ();
				group->add (name, value);
//log ("\n    Added key \"%s\" value \"%s\" to the map %d...", name, value, mapCounter);
				break;}
			case '{': {
//log ("\n    Reading side planes...");
				WorldcraftSidePlaneCollection *sidePlanes = readRestOfSidePlanes ();
				Object *object = new Object (convertToFaces (sidePlanes));
				objects->push_back (object);
//log ("\n    Added key \"%s\" value \"%s\" to the map %d...", "faces", "A FACE COLLECTION", mapCounter);
				break;}
			case '}': {
				return group;}
			default:
				quit ("Unexpected input while parsing a worldcraft group.");
		}
	}
    quit ("Unexpected end of file without finding \"}\".");
}

char *readRestOfDoubleQuotedString () {
	//Does NOT treat the escape character specially since it is in many worldcraft strings as a
	//file delimiter. No internal double quotes are possible since it stops reading on first encounter.
	char buffer [1000]; char *start = &buffer [0]; char *next = start - 1; //Should be plenty for worldcraft.

	*++next = file.get ();
	while (!file.eof ()) {
		switch (*next) {
			case '"': {
				//End of string. All done...
				*next = '\0'; //end of string character (added where get failed).
				char *space = new char [next - start + 1]; 
				strcpy (space, start);
				return space;}
			default:
				//Already stored in buffer, carry on...
				*++next = file.get ();
		}
	}
	*next = '\0'; //end of string character (added where get failed).
	quit ("Unexpected end of file while reading double quoted string \"%s\".", start);
}

WorldcraftSidePlaneCollection *readRestOfSidePlanes () {
    //Reads a bunch of worldcraft sidePlane packets.

	WorldcraftSidePlaneCollection *sidePlanes = new WorldcraftSidePlaneCollection;
	char next;
	while (file >> next) {//>> skips whitespace
		switch (next) {
			case '(':
				sidePlanes->push_back (readRestOfSidePlane ());
				break;
			case '}':
				return sidePlanes;
			default:
				quit ("Unexpected input while parsing a worldcraft brush's side plane; i.e., (...");
		}
	}
    quit ("Unexpected end of file without finding \")\" to match \"(\".");
}

char *doubleProblem = "Unexpected error while reading brush's side plane packed of form (..) (..) (..) ...";
char *leftParenthesisProblem = "Expected \"(\".";
char *rightParenthesisProblem = "Expected \")\".";
char *leftSquareBracketProblem = "Expected \"[\".";
char *rightSquareBracketProblem = "Expected \"]\".";

WorldcraftSidePlane *readRestOfSidePlane () {
    //Reads a side plane packet; 
	//	e.g. ( 176 -384 400 ) ( 176 -404 400 ) ( 429 -404 400 ) BRICK1 [ 1 0 0 -1 ] [ 0 -1 0 -2 ] 0 0.3 0.3 
	//where the first parenthesis has already been removed.

	WorldcraftSidePlane *sidePlane = new WorldcraftSidePlane;

	//Pick up point1.
	char next;
	if (!(file >> sidePlane->point1.x)) quit (doubleProblem);
	if (!(file >> sidePlane->point1.y)) quit (doubleProblem);
	if (!(file >> sidePlane->point1.z)) quit (doubleProblem);
	if (!(file >> next) || next != ')') quit (rightParenthesisProblem);

	//Pick up point2.
	if (!(file >> next) || next != '(') quit (leftParenthesisProblem);
	if (!(file >> sidePlane->point2.x)) quit (doubleProblem);
	if (!(file >> sidePlane->point2.y)) quit (doubleProblem);
	if (!(file >> sidePlane->point2.z)) quit (doubleProblem);
	if (!(file >> next) || next != ')') quit (rightParenthesisProblem);

	//Pick up point3.
	if (!(file >> next) || next != '(') quit (leftParenthesisProblem);
	if (!(file >> sidePlane->point3.x)) quit (doubleProblem);
	if (!(file >> sidePlane->point3.y)) quit (doubleProblem);
	if (!(file >> sidePlane->point3.z)) quit (doubleProblem);
	if (!(file >> next) || next != ')') quit (rightParenthesisProblem);

	//Pick up the texture name.
	sidePlane->textureName = readTextureName ();

	//Pick up the x-axis.
	if (!(file >> next) || next != '[') quit (leftSquareBracketProblem);
	if (!(file >> sidePlane->xAxis.x)) quit (doubleProblem);
	if (!(file >> sidePlane->xAxis.y)) quit (doubleProblem);
	if (!(file >> sidePlane->xAxis.z)) quit (doubleProblem);
	if (!(file >> sidePlane->xAxisOffset)) quit (doubleProblem);
	if (!(file >> next) || next != ']') quit (rightSquareBracketProblem);

	//Pick up the y-axis.
	if (!(file >> next) || next != '[') quit (leftSquareBracketProblem);
	if (!(file >> sidePlane->yAxis.x)) quit (doubleProblem);
	if (!(file >> sidePlane->yAxis.y)) quit (doubleProblem);
	if (!(file >> sidePlane->yAxis.z)) quit (doubleProblem);
	if (!(file >> sidePlane->yAxisOffset)) quit (doubleProblem);
	if (!(file >> next) || next != ']') quit (rightSquareBracketProblem);

    //Pick up the rotation"
	if (!(file >> sidePlane->rotation)) quit (doubleProblem);

    //Pick up the x axis scale"
	if (!(file >> sidePlane->xAxisScale)) quit (doubleProblem);

    //Pick up the y axis scale"
	if (!(file >> sidePlane->yAxisScale)) quit (doubleProblem);

	//All done.
	return sidePlane;
}

char *readTextureName () {
	//Does NOT treat the escape character specially since it is in many worldcraft strings as a
	//file delimiter. No internal double quotes are possible since it stops reading on first encounter.
	char buffer [1000]; char *start = &buffer [0]; char *next = start - 1; //Should be plenty for worldcraft.

	char nextCharacter;
	if (!(file >> nextCharacter)) quit ("End of file while reading texture name.");
	*++next = nextCharacter;
	while (!file.eof ()) {
		switch (*next) {
			case ' ':
			case '\t':
			case '\n':
			case '\f': {
				//End of string. All done...
				*next = '\0'; //end of string character (added where get failed).
				char *space = new char [next - start + 1]; 
				strcpy (space, start);
				return space;}
			default:
				//Already stored in buffer, carry on...
				*++next = file.get ();
		}
	}
	*next = '\0'; //end of string character (added where get failed).
	quit ("Unexpected end of file while reading texture name \"%s\".", start);
}

PointCollection *cutUpBy (PointCollection *points, WorldcraftSidePlaneCollection *sidePlanes, long indexOfExcludedPlane) {
//log ("\nEnter cut up with %d points.", points->size ());
//for (long i = 0; i < points->size (); i++) {
//log ("\n   [%3.2f,%3.2f,%3.2f].", ((*points)[i])->x, ((*points)[i])->y, ((*points)[i])->z);
//}
	PointCollection *oldPoints = points; //Don't delete points.
	for (long planeIndex = 0; planeIndex < sidePlanes->size (); planeIndex++) {
//log ("\nConsider plane [%3.2f,%3.2f,%3.2f,%3.2f].", plane.normal.x, plane.normal.y, plane.normal.z, plane.minusP0DotNormal);
//if (pointIndex == indexOfExcludedPlane) log ("\nEXCLUDE THIS PLANE...");
		if (planeIndex != indexOfExcludedPlane) {
			WorldcraftSidePlane *sidePlane = (*sidePlanes) [planeIndex];
			Plane &plane = sidePlane->plane;
			Plane flippedPlane = Plane (-plane.normal, -plane.minusP0DotNormal);
			PointCollection *newPoints = clip (flippedPlane, oldPoints);
//log ("\nClipped to %d points.", newPoints->size ());
//for (i = 0; i < newPoints->size (); i++) {
//log ("\n   [%3.2f,%3.2f,%3.2f].", ((*newPoints)[i])->x, ((*newPoints)[i])->y, ((*newPoints)[i])->z);
//}
			if (oldPoints != points) deletePointCollection (oldPoints);
			oldPoints = newPoints;
		}
	}
//log ("\nReturning %d points.", oldPoints->size ());
//for (i = 0; i < oldPoints->size (); i++) {
//log ("\n   [%3.2f,%3.2f,%3.2f].", ((*oldPoints)[i])->x, ((*oldPoints)[i])->y, ((*oldPoints)[i])->z);
//}
	return oldPoints;
}

PointCollection *enlarge (PointCollection *points, Plane &plane) {
	//Find the center (just the average).
	Point sum; sum.zero ();
	for (long pointIndex = 0; pointIndex < points->size (); pointIndex++) {
		sum += *((*points) [pointIndex]);
	}
	Point center = sum / (double) points->size ();

	//Move each point out by doubling it's distance to the center and project onto the plance
	//to make sure is right on the plane (to make sure numerical inaccuracies don't creep in).
	PointCollection *newPoints = new PointCollection;
	for (pointIndex = 0; pointIndex < points->size (); pointIndex++) {
		Point *point = (*points) [pointIndex];
		Point outwardMovedPoint = plane.projectionOfPoint (center + (*point - center) * 2.0);
		newPoints->push_back (outwardMovedPoint.newCopy ());
	}
	return newPoints;
}

double inline worldcraftToMeters () {
    //Since 0.0254 m/in * 12in/ft = 0.3048 m/ft / 8wcu/ft = 0.0381 m/wcu (worldcraft units),
    //multiply by 0.0381 to convert to meters."
    return 0.0381;
}

FaceCollection *convertToFaces (WorldcraftSidePlaneCollection *sidePlanes) {
	//Convert the side planes into faces (one face per side plane unless a particular side plane is
	//degenerate; i.e., consists of a zero-area face).
//log ("\n\nConvert object with %d side planes...", sidePlanes->size ());
	FaceCollection *faces = new FaceCollection;

	//Set up 2 new collections and iterators: planes and transformations.
	PlaneCollection *planes = new PlaneCollection;
	TransformationCollection *transformations = new TransformationCollection;

	//Get the point and the texture mapping transformation for each side point by iterating over the side planes.
	for (long sidePlaneIndex = 0; sidePlaneIndex < sidePlanes->size (); sidePlaneIndex++) {

		//Consider this side plane.
		WorldcraftSidePlane &sidePlane = *((*sidePlanes) [sidePlaneIndex]);

		//Convert points to a standard coordinate system (worldcraft uses x +ve is right, y+ve is back, z+ve is up).
		//Also convert from clockwise order p1,p2,p3 to counter-clockwise order p2,p1,p3.
		//Note: these are NOT points in a face but merely 3 points which can be used to compute the plane of the face.
		Point point1 = {sidePlane.point1.x, sidePlane.point1.z, -sidePlane.point1.y};
		Point point2 = {sidePlane.point2.x, sidePlane.point2.z, -sidePlane.point2.y};
		Point point3 = {sidePlane.point3.x, sidePlane.point3.z, -sidePlane.point3.y};
//log ("\n   Original point1 for plane [%3.2f,%3.2f,%3.2f].", sidePlane.point1.x, sidePlane.point1.y, sidePlane.point1.z);
//log ("\n   Original point2 for plane [%3.2f,%3.2f,%3.2f].", sidePlane.point2.x, sidePlane.point2.y, sidePlane.point2.z);
//log ("\n   Original point3 for plane [%3.2f,%3.2f,%3.2f].", sidePlane.point3.x, sidePlane.point3.y, sidePlane.point3.z);

//log ("\n   Point1 for plane [%3.2f,%3.2f,%3.2f].", point1.x, point1.y, point1.z);
//log ("\n   Point2 for plane [%3.2f,%3.2f,%3.2f].", point2.x, point2.y, point2.z);
//log ("\n   Point3 for plane [%3.2f,%3.2f,%3.2f].", point3.x, point3.y, point3.z);
		
		//Compute the plane and carry on only if it's a legitimate plane (Yup... Worldcraft can give you junk).
		Plane *plane = Plane::fromDangerousPoints (point2, point1, point3);
		if (plane == NULL) {
			log ("\nDiscarded a zero-area triangle; points are \n\t%3.2f,%3.2f,%3.2f]\n\t%3.2f,%3.2f,%3.2f]\n\t%3.2f,%3.2f,%3.2f]",
				point1.x, point1.y, point1.z, point2.x, point2.y, point2.z, point3.x, point3.y, point3.z);
			static bool thisIsFirstError = true;
			if (thisIsFirstError) {
				prompt ("\nDiscarded a zero-area triangle; points are \n\t%3.2f,%3.2f,%3.2f]\n\t%3.2f,%3.2f,%3.2f]\n\t%3.2f,%3.2f,%3.2f]",
					point1.x, point1.y, point1.z, point2.x, point2.y, point2.z, point3.x, point3.y, point3.z);
				thisIsFirstError = false;
			}
		} else {
			//Just record the plane.
//log ("\n   Plane [%3.2f,%3.2f,%3.2f,%3.2f].", plane->normal.x, plane->normal.y, plane->normal.z, plane->minusP0DotNormal);
			sidePlane.plane = *plane;

			//Recall: worldcraft coordinates are of the form: x z -y.
			//Convert the worldcraft axes for texture mapping to the standard system.
			//These can be used as rows in a matrix that gives us R.
			Point xAxis = {sidePlane.xAxis.x, sidePlane.xAxis.z, -sidePlane.xAxis.y};
			Point yAxis = {sidePlane.yAxis.x, sidePlane.yAxis.z, -sidePlane.yAxis.y};
			Point zAxis = xAxis.cross (yAxis);
//log ("\n   xAxis [%3.2f,%3.2f,%3.2f,%3.2f].", xAxis.x, xAxis.y, xAxis.z);
//log ("\n   yAxis [%3.2f,%3.2f,%3.2f,%3.2f].", yAxis.x, yAxis.y, yAxis.z);
//log ("\n   zAxis [%3.2f,%3.2f,%3.2f,%3.2f].", zAxis.x, zAxis.y, zAxis.z);
			//Replace by the inverse (the transpose is the inverse in this case). 
			//When these rows are placed in a transformation matrix, it gives us R-1.
			Point uAxis = {xAxis.x, yAxis.x, zAxis.x};
			Point vAxis = {xAxis.y, yAxis.y, zAxis.y};
			Point wAxis = {xAxis.z, yAxis.z, zAxis.z};

			//Read the texture to get the extent (we assume it's in directory Texture)....
			char *bmpFilePathName [500];
			strcpy ((char *) &bmpFilePathName [0], directoryName);
			strcat ((char *) &bmpFilePathName [0], "Textures\\");
			strcat ((char *) &bmpFilePathName [0], sidePlane.textureName);
			strcat ((char *) &bmpFilePathName [0], ".bmp");
			char *newName = (char *) &bmpFilePathName [0];
			long textureWidth; long textureHeight;
			bool readWorked = true;
			if (!Texture::readTextureExtent (newName, textureWidth, textureHeight)) {
				//Try again with .tga suffix.
				char *tgaFilePathName [500];
				strcpy ((char *) &tgaFilePathName [0], directoryName);
				strcat ((char *) &tgaFilePathName [0], "Textures\\");
				strcat ((char *) &tgaFilePathName [0], sidePlane.textureName);
				strcat ((char *) &tgaFilePathName [0], ".tga");
				newName = (char *) &tgaFilePathName [0];
				if (!Texture::readTextureExtent (newName, textureWidth, textureHeight)) {
					//Give up and give a .bmp file error message.
					log ("\nCould not find texture \"%s\".", &bmpFilePathName [0]);
					prompt ("\nCould not find texture \"%s\".", &bmpFilePathName [0]);
					textureWidth = 1; textureHeight = 1; readWorked = false;
				}
			}
			if (readWorked) {//Remember the new full path name.
				delete [] sidePlane.textureName; 
				sidePlane.textureName = new char [strlen (newName) + 1];
				strcpy (sidePlane.textureName, newName);
			}

//log ("\n   extent [%3.2f,%3.2f].", textureWidth*1.0, textureHeight*1.0);
//log ("\n   axis scale [%3.2f,%3.2f].", sidePlane.xAxisScale, sidePlane.yAxisScale);
//log ("\n   axis offset [%3.2f,%3.2f].", sidePlane.xAxisOffset, sidePlane.yAxisOffset);
			//Create and record a transformation to map texture coordinates from normal coordinates.
			Point axisScale = {1.0/sidePlane.xAxisScale, -1.0/sidePlane.yAxisScale, 1.0};
			Point textureTranslation = {sidePlane.xAxisOffset, -sidePlane.yAxisOffset, 0.0};
			Point textureScale = {1.0/textureWidth, 1.0/textureHeight, 1.0};
//log ("\n   New scale is [%3.2f,%3.2f,%3.2f].", axisScale.x, axisScale.y, axisScale.z);
			
			Transformation &transformation = sidePlane.transformation;
			transformation.rotateToAxes (uAxis, vAxis, wAxis); //R-1
//log ("\nRotation:"); transformation.show ();
			transformation.postScaleBy (axisScale); //S-1
//log ("\nScale:"); transformation.show ();
			transformation.postTranslateBy (textureTranslation); //T-1
//log ("\nTranslate:"); transformation.show ();
			transformation.postScaleBy (textureScale); //N-1
//log ("\nTexture Scale:"); transformation.show ();
		}
	}

	//Compute the face coordinates by starting with the points of a very large face and cutting it up
	//by ALL the planes. The resulting points are the coordinates of the faces we wish to build. Once 
	//done, compute the texture coordinates using the transformation computed earlier.
	for (sidePlaneIndex = 0; sidePlaneIndex < sidePlanes->size (); sidePlaneIndex++) {

		//Consider this side plane and it's associated heavily used parts.
		WorldcraftSidePlane &sidePlane = *((*sidePlanes) [sidePlaneIndex]);
		Plane &plane = sidePlane.plane;
		Transformation &transformation = sidePlane.transformation;

		//Construct the points for the face coordinates
		PointCollection *largeFacePoints = anyFarCornerPoints (plane);
//for (long i = 0; i < largeFacePoints->size (); i++) {
//log ("\n   Far point [%3.2f,%3.2f,%3.2f].", ((*largeFacePoints)[i])->x, ((*largeFacePoints)[i])->y, ((*largeFacePoints)[i])->z);
//}
//log ("\nLarge face points size is %d.", largeFacePoints->size ());
//for (long i = 0; i < largeFacePoints->size (); i++) {
//log ("\n   [%3.2f,%3.2f,%3.2f].", ((*largeFacePoints)[i])->x, ((*largeFacePoints)[i])->y, ((*largeFacePoints)[i])->z);
//}
		PointCollection *cutUpFacePoints = cutUpBy (largeFacePoints, sidePlanes, sidePlaneIndex);
//log ("\nCutup face points size is %d.", cutUpFacePoints->size ());
//for (i = 0; i < cutUpFacePoints->size (); i++) {
//log ("\n   [%3.2f,%3.2f,%3.2f].", ((*cutUpFacePoints)[i])->x, ((*cutUpFacePoints)[i])->y, ((*cutUpFacePoints)[i])->z);
//}
		PointCollection *enlargedFacePoints = enlarge (cutUpFacePoints, plane);
//log ("\nEnlarged face points size is %d.", enlargedFacePoints->size ());
//for (i = 0; i < enlargedFacePoints->size (); i++) {
//log ("\n   [%3.2f,%3.2f,%3.2f].", ((*enlargedFacePoints)[i])->x, ((*enlargedFacePoints)[i])->y, ((*enlargedFacePoints)[i])->z);
//}
		PointCollection *moreAccurateCutUpFacePoints = cutUpBy (enlargedFacePoints, sidePlanes, sidePlaneIndex);
//log ("\nAccurate face points size is %d.", moreAccurateCutUpFacePoints->size ());
//for (i = 0; i < moreAccurateCutUpFacePoints->size (); i++) {
//log ("\n   [%3.2f,%3.2f,%3.2f].", ((*moreAccurateCutUpFacePoints)[i])->x, ((*moreAccurateCutUpFacePoints)[i])->y, ((*moreAccurateCutUpFacePoints)[i])->z);
//}
		deletePointCollection (largeFacePoints); deletePointCollection (cutUpFacePoints);
		deletePointCollection (enlargedFacePoints);

		//Finally, replace standard 3D points by game points.
		GamePointCollection *gamePoints = new GamePointCollection;
//log ("\nFace:");
		for (long index = 0; index < moreAccurateCutUpFacePoints->size (); index++) {
			Point *worldcraftPoint = (*moreAccurateCutUpFacePoints) [index];
			Point coordinatePoint = *worldcraftPoint * worldcraftToMeters ();
			Point texturePoint = *worldcraftPoint * transformation;

			GamePoint *gamePoint = new GamePoint;
			gamePoint->x = coordinatePoint.x;
			gamePoint->y = coordinatePoint.y;
			gamePoint->z = coordinatePoint.z;
			gamePoint->nx = plane.normal.x;
			gamePoint->ny = plane.normal.y;
			gamePoint->nz = plane.normal.z;
			gamePoint->tx = texturePoint.x;
			gamePoint->ty = texturePoint.y;

			gamePoints->push_back (gamePoint);
/*
log ("\n   p = [%3.2f,%3.2f,%3.2f], n = [%3.2f,%3.2f,%3.2f], t = [%3.2f,%3.2f].",
gamePoint->x,
gamePoint->y,
gamePoint->z,
gamePoint->nx,
gamePoint->ny,
gamePoint->nz,
gamePoint->tx,
gamePoint->ty);
*/
		}
		deletePointCollection (moreAccurateCutUpFacePoints);

		//Create a new face if there are sufficient points.
//log ("\nGame points size is %d.", gamePoints->size ());
		if (gamePoints->empty ()) {
			deleteGamePointCollection (gamePoints);
			delete [] sidePlane.textureName;
//log ("\nNO FACE BUILT...");
		} else {
			Face *face = new Face (plane, gamePoints, sidePlane.textureName);
			faces->push_back (face);
//log ("\nFACE BUILT...");
		}
	}

//log (" => %d faces contructed...", faces->size ());
	return faces;
}

